@@ -30,7 +30,6 @@
 #include "json-ipc.h"
 #include "ctl-client.h"
 #include "ctl-commands.h"
-#include "ctl-server.h"
 #include "strlcpy.h"
 #include "util.h"
 #include "option-parser.h"
@@ -228,7 +227,7 @@ static json_t* json_from_buffer(struct ctl_client* self)
 	}
 
 	json_error_t err;
-	json_t* root = json_loadb(self->read_buffer, self->read_len, 0, &err);
+	json_t* root = json_loadb(self->read_buffer, self->read_len, JSON_DISABLE_EOF_CHECK, &err);
 	if (root) {
 		advance_read_buffer(&self->read_buffer, &self->read_len,
 				err.position);
@@ -619,14 +618,26 @@ static int ctl_client_register_for_events(struct ctl_client* self,
 }
 
 static int ctl_client_reconnect_event_loop(struct ctl_client* self,
-		struct jsonipc_request* request, int timeout)
+		struct jsonipc_request* request)
 {
-	if (ctl_client_connect(self, timeout) != 0)
+	if (ctl_client_connect(self, -1) != 0)
 		return -1;
 
 	return ctl_client_register_for_events(self, request);
 }
 
+static int block_until_reconnect(struct ctl_client* self,
+		struct jsonipc_request* request)
+{
+	while (ctl_client_reconnect_event_loop(self, request) != 0)
+		if (usleep(50000) == -1) {
+			DEBUG("Interrupted waiting for the IPC socket");
+			return -1;
+		}
+
+	return 0;
+}
+
 static int ctl_client_event_loop(struct ctl_client* self,
 		struct jsonipc_request* request)
 {
@@ -643,8 +654,8 @@ static int ctl_client_event_loop(struct ctl_client* self,
 			if (errno == ECONNRESET) {
 				send_shutdown_event(self);
 				if (self->flags & CTL_CLIENT_RECONNECT &&
-						ctl_client_reconnect_event_loop(
-							self, request, -1) == 0)
+						block_until_reconnect(
+							self, request) == 0)
 					continue;
 			}
 			break;
@@ -660,12 +671,16 @@ static int ctl_client_event_loop(struct ctl_client* self,
 }
 
 static int ctl_client_print_single_command(struct ctl_client* self,
-		struct jsonipc_request* request)
+		enum cmd_type cmd, struct jsonipc_request* request)
 {
 	struct jsonipc_response* response = ctl_client_run_single_command(self,
 			request);
-	if (!response)
+	if (!response) {
+		if (errno == ECONNRESET && cmd == CMD_WAYVNC_EXIT)
+			return 0;
+
 		return 1;
+	}
 
 	int result = ctl_client_print_response(self, request, response);
 	jsonipc_response_destroy(response);
@@ -897,7 +912,7 @@ int ctl_client_run_command(struct ctl_client* self,
 		result = ctl_client_event_loop(self, request);
 		break;
 	default:
-		result = ctl_client_print_single_command(self, request);
+		result = ctl_client_print_single_command(self, cmd, request);
 		break;
 	}
 
