@@ -328,24 +328,54 @@ static struct cmd_response* generate_version_object()
 	return response;
 }
 
+static struct ctl_server_client* ctl_server_client_first(struct ctl* self)
+{
+	return self->actions.client_next(self, NULL);
+}
+
+static struct ctl_server_client* ctl_server_client_next(struct ctl* self,
+		struct ctl_server_client* prev)
+{
+	return self->actions.client_next(self, prev);
+}
+
+static void ctl_server_client_get_info(struct ctl* self,
+		const struct ctl_server_client* client,
+		struct ctl_server_client_info* info)
+{
+	return self->actions.client_info(client, info);
+}
+
 static struct cmd_response* generate_vnc_client_list(struct ctl* self)
 {
-	struct ctl_server_vnc_client* clients;
-	size_t num_clients = self->actions.get_client_list(self, &clients);
 	struct cmd_response* response = cmd_ok();
-
 	response->data = json_array();
-	for (size_t i = 0; i < num_clients; ++i) {
-		json_t* packed = json_pack("{s:s}", "id", clients[i].id);
-		if (clients[i].hostname[0] != '\0')
+
+	struct ctl_server_client* client;
+	for (client = ctl_server_client_first(self); client;
+			client = ctl_server_client_next(self, client)) {
+		struct ctl_server_client_info info = {};
+		ctl_server_client_get_info(self, client, &info);
+
+		char id_str[64];
+		snprintf(id_str, sizeof(id_str), "%d", info.id);
+		json_t* packed = json_pack("{s:s}", "id", id_str);
+
+		if (info.hostname)
 			json_object_set_new(packed, "hostname",
-					json_string(clients[i].hostname));
-		if (clients[i].username[0] != '\0')
+					json_string(info.hostname));
+
+		if (info.username)
 			json_object_set_new(packed, "username",
-					json_string(clients[i].username));
+					json_string(info.username));
+
+		if (info.seat)
+			json_object_set_new(packed, "seat",
+					json_string(info.seat));
+
 		json_array_append_new(response->data, packed);
 	}
-	free(clients);
+
 	return response;
 }
 
@@ -829,15 +859,18 @@ struct cmd_response* cmd_failed(const char* fmt, ...)
 }
 
 json_t* pack_connection_event_params(
-		const char* client_id,
-		const char* client_hostname,
-		const char* client_username,
+		const struct ctl_server_client_info *info,
 		int new_connection_count)
 {
-	return json_pack("{s:s, s:s?, s:s?, s:i}",
-			"id", client_id,
-			"hostname", client_hostname,
-			"username", client_username,
+	// TODO: Why is the id a string?
+	char id_str[64];
+	snprintf(id_str, sizeof(id_str), "%d", info->id);
+
+	return json_pack("{s:s, s:s?, s:s?, s:s?, s:i}",
+			"id", id_str,
+			"hostname", info->hostname,
+			"username", info->username,
+			"seat", info->seat,
 			"connection_count", new_connection_count);
 }
 
@@ -879,34 +912,28 @@ int ctl_server_enqueue_event(struct ctl* self, enum event_type evt_type,
 
 static void ctl_server_event_connect(struct ctl* self,
 		enum event_type evt_type,
-		const char* client_id,
-		const char* client_hostname,
-		const char* client_username,
+		const struct ctl_server_client_info *info,
 		int new_connection_count)
 {
-	json_t* params = pack_connection_event_params(client_id, client_hostname,
-			client_username, new_connection_count);
+	json_t* params =
+		pack_connection_event_params(info, new_connection_count);
 	ctl_server_enqueue_event(self, evt_type, params);
 }
 
 void ctl_server_event_connected(struct ctl* self,
-		const char* client_id,
-		const char* client_hostname,
-		const char* client_username,
+		const struct ctl_server_client_info *info,
 		int new_connection_count)
 {
-	ctl_server_event_connect(self, EVT_CLIENT_CONNECTED, client_id,
-			client_hostname, client_username, new_connection_count);
+	ctl_server_event_connect(self, EVT_CLIENT_CONNECTED, info,
+			new_connection_count);
 }
 
 void ctl_server_event_disconnected(struct ctl* self,
-		const char* client_id,
-		const char* client_hostname,
-		const char* client_username,
+		const struct ctl_server_client_info *info,
 		int new_connection_count)
 {
-	ctl_server_event_connect(self, EVT_CLIENT_DISCONNECTED, client_id,
-			client_hostname, client_username, new_connection_count);
+	ctl_server_event_connect(self, EVT_CLIENT_DISCONNECTED, info,
+			new_connection_count);
 }
 
 void ctl_server_event_capture_changed(struct ctl* self,
