@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2019 - 2022 Andri Yngvason
+ * Copyright (c) 2019 - 2023 Andri Yngvason
  *
  * Permission to use, copy, modify, and/or distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -71,6 +71,12 @@
 
 #define MAYBE_UNUSED __attribute__((unused))
 
+enum socket_type {
+	SOCKET_TYPE_TCP = 0,
+	SOCKET_TYPE_UNIX,
+	SOCKET_TYPE_WEBSOCKET,
+};
+
 struct wayvnc {
 	bool do_exit;
 
@@ -85,7 +91,7 @@ struct wayvnc {
 	struct zwlr_data_control_manager_v1* data_control_manager;
 
 	struct output* selected_output;
-	const struct seat* selected_seat;
+	struct seat* selected_seat;
 
 	struct screencopy screencopy;
 
@@ -116,6 +122,8 @@ struct wayvnc_client {
 	struct wayvnc* server;
 	struct nvnc_client* nvnc_client;
 
+	struct seat* seat;
+
 	unsigned id;
 	struct pointer pointer;
 	struct keyboard keyboard;
@@ -128,6 +136,7 @@ static void on_nvnc_client_new(struct nvnc_client* client);
 void switch_to_output(struct wayvnc*, struct output*);
 void switch_to_next_output(struct wayvnc*);
 void switch_to_prev_output(struct wayvnc*);
+static void client_init_seat(struct wayvnc_client* self);
 static void client_init_pointer(struct wayvnc_client* self);
 static void client_init_keyboard(struct wayvnc_client* self);
 static void client_init_data_control(struct wayvnc_client* self);
@@ -155,7 +164,7 @@ static bool registry_add_input(void* data, struct wl_registry* registry,
 
 		struct seat* seat = seat_new(wl_seat, id);
 		if (!seat) {
-			wl_seat_destroy(wl_seat);
+			wl_seat_release(wl_seat);
 			return true;
 		}
 
@@ -252,6 +261,20 @@ static void registry_add(void* data, struct wl_registry* registry,
 		return;
 }
 
+static void disconnect_seat_clients(struct wayvnc* self, struct seat* seat)
+{
+	struct nvnc_client* nvnc_client;
+	for (nvnc_client = nvnc_client_first(self->nvnc); nvnc_client;
+			nvnc_client = nvnc_client_next(nvnc_client)) {
+		struct wayvnc_client* client = nvnc_get_userdata(nvnc_client);
+		assert(client);
+
+		if (client->seat == seat) {
+			nvnc_client_close(nvnc_client);
+		}
+	}
+}
+
 static void registry_remove(void* data, struct wl_registry* registry,
 			    uint32_t id)
 {
@@ -280,14 +303,9 @@ static void registry_remove(void* data, struct wl_registry* registry,
 	struct seat* seat = seat_find_by_id(&self->seats, id);
 	if (seat) {
 		nvnc_log(NVNC_LOG_INFO, "Seat %s went away", seat->name);
+		disconnect_seat_clients(self, seat);
 		wl_list_remove(&seat->link);
 		seat_destroy(seat);
-
-		if (seat == self->selected_seat) {
-			nvnc_log(NVNC_LOG_ERROR, "Selected seat went away. Exiting...");
-			wayvnc_exit(self);
-		}
-
 		return;
 	}
 }
@@ -489,33 +507,27 @@ struct cmd_response* on_output_switch(struct ctl* ctl,
 	return cmd_ok();
 }
 
-static int get_client_list(struct ctl* ctl,
-		struct ctl_server_vnc_client** clients)
+static struct ctl_server_client *client_next(struct ctl* ctl,
+		struct ctl_server_client *prev)
 {
 	struct wayvnc* self = ctl_server_userdata(ctl);
-	if (self->nr_clients == 0) {
-		*clients = NULL;
-		return 0;
-	}
-	*clients = calloc(self->nr_clients, sizeof(**clients));
-	struct nvnc_client* nvnc_client = nvnc_client_first(self->nvnc);
-	for (int i = 0; i < self->nr_clients && nvnc_client; ++i) {
-		struct wayvnc_client* client = nvnc_get_userdata(nvnc_client);
-		struct ctl_server_vnc_client* ctl_client =&(*clients)[i];
+	struct nvnc_client* vnc_prev = (struct nvnc_client*)prev;
 
-		snprintf(ctl_client->id, sizeof(ctl_client->id), "%u",
-				client->id);
-		const char* hostname = nvnc_client_get_hostname(nvnc_client);
-		if (hostname)
-			strlcpy(ctl_client->hostname, hostname,
-					sizeof(ctl_client->hostname));
-		const char* username = nvnc_client_get_auth_username(nvnc_client);
-		if (username)
-			strlcpy(ctl_client->username, username,
-					sizeof(ctl_client->username));
-		nvnc_client = nvnc_client_next(nvnc_client);
-	}
-	return self->nr_clients;
+	return prev ? (struct ctl_server_client*)nvnc_client_next(vnc_prev) :
+		(struct ctl_server_client*)nvnc_client_first(self->nvnc);
+}
+
+static void client_info(const struct ctl_server_client* client_handle,
+		struct ctl_server_client_info* info)
+{
+	const struct nvnc_client *vnc_client =
+		(const struct nvnc_client*)client_handle;
+	const struct wayvnc_client *client = nvnc_get_userdata(vnc_client);
+
+	info->id = client->id;
+	info->hostname = nvnc_client_get_hostname(vnc_client);
+	info->username = nvnc_client_get_auth_username(vnc_client);
+	info->seat = client->seat ? client->seat->name : NULL;
 }
 
 static int get_output_list(struct ctl* ctl,
@@ -693,14 +705,38 @@ static int blank_screen(struct wayvnc* self)
 	return 0;
 }
 
-int init_nvnc(struct wayvnc* self, const char* addr, uint16_t port, bool is_unix)
+static char* get_cfg_path(const struct cfg* cfg, char* dst, const char* src)
 {
-	self->nvnc = is_unix ? nvnc_open_unix(addr) : nvnc_open(addr, port);
+	if (!cfg->use_relative_paths || src[0] == '/') {
+		strlcpy(dst, src, PATH_MAX);
+		return dst;
+	}
+
+	snprintf(dst, PATH_MAX, "%s/%s", cfg->directory, src);
+	return dst;
+}
+
+static int init_nvnc(struct wayvnc* self, const char* addr, uint16_t port,
+		enum socket_type socket_type)
+{
+	switch (socket_type) {
+		case SOCKET_TYPE_TCP:
+			self->nvnc = nvnc_open(addr, port);
+			break;
+		case SOCKET_TYPE_UNIX:
+			self->nvnc = nvnc_open_unix(addr);
+			break;
+		case SOCKET_TYPE_WEBSOCKET:
+			self->nvnc = nvnc_open_websocket(addr, port);
+			break;
+		default:
+			abort();
+	}
 	if (!self->nvnc) {
 		nvnc_log(NVNC_LOG_ERROR, "Failed to bind to address. Add -Ldebug to the argument list for more info.");
 		return -1;
 	}
-	if (is_unix)
+	if (socket_type == SOCKET_TYPE_UNIX)
 		nvnc_log(NVNC_LOG_INFO, "Listening for connections on %s", addr);
 	else
 		nvnc_log(NVNC_LOG_INFO, "Listening for connections on %s:%d", addr, port);
@@ -715,11 +751,45 @@ int init_nvnc(struct wayvnc* self, const char* addr, uint16_t port, bool is_unix
 
 	nvnc_set_name(self->nvnc, "WayVNC");
 
-	if (self->cfg.enable_auth &&
-	    nvnc_enable_auth(self->nvnc, self->cfg.private_key_file,
-	                     self->cfg.certificate_file, on_auth, self) < 0) {
-		nvnc_log(NVNC_LOG_ERROR, "Failed to enable authentication");
-		goto failure;
+	enum nvnc_auth_flags auth_flags = 0;
+	if (self->cfg.enable_auth) {
+		auth_flags |= NVNC_AUTH_REQUIRE_AUTH;
+	}
+	if (!self->cfg.relax_encryption) {
+		auth_flags |= NVNC_AUTH_REQUIRE_ENCRYPTION;
+	}
+
+	if (self->cfg.enable_auth) {
+		if (nvnc_enable_auth(self->nvnc, auth_flags, on_auth, self) < 0) {
+			nvnc_log(NVNC_LOG_ERROR, "Failed to enable authentication");
+			goto failure;
+		}
+
+		if (self->cfg.rsa_private_key_file) {
+			char tmp[PATH_MAX];
+			const char* key_file = get_cfg_path(&self->cfg, tmp,
+					self->cfg.rsa_private_key_file);
+			if (nvnc_set_rsa_creds(self->nvnc, key_file) < 0) {
+				nvnc_log(NVNC_LOG_ERROR, "Failed to load RSA credentials");
+				goto failure;
+			}
+		}
+
+		if (self->cfg.private_key_file) {
+			char key_file[PATH_MAX];
+			char cert_file[PATH_MAX];
+
+			get_cfg_path(&self->cfg, key_file,
+					self->cfg.private_key_file);
+			get_cfg_path(&self->cfg, cert_file,
+					self->cfg.certificate_file);
+			int r = nvnc_set_tls_creds(self->nvnc, key_file,
+					cert_file);
+			if (r < 0) {
+				nvnc_log(NVNC_LOG_ERROR, "Failed to enable TLS authentication");
+				goto failure;
+			}
+		}
 	}
 
 	if (self->pointer_manager)
@@ -878,6 +948,9 @@ void wayvnc_process_frame(struct wayvnc* self)
 	nvnc_fb_set_transform(buffer->nvnc_fb,
 			(enum nvnc_transform)buffer_transform);
 
+	pixman_region_intersect_rect(&damage, &damage, 0, 0, buffer->width,
+			buffer->height);
+
 	nvnc_display_feed_buffer(self->nvnc_display, buffer->nvnc_fb,
 			&damage);
 
@@ -929,18 +1002,14 @@ int check_cfg_sanity(struct cfg* cfg)
 
 		if (!nvnc_has_auth()) {
 			nvnc_log(NVNC_LOG_ERROR, "Authentication can't be enabled because it was not selected during build");
-			return -1;
-		}
-
-		if (!cfg->certificate_file) {
-			nvnc_log(NVNC_LOG_ERROR, "Authentication enabled, but missing certificate_file");
 			rc = -1;
 		}
 
-		if (!cfg->private_key_file) {
-			nvnc_log(NVNC_LOG_ERROR, "Authentication enabled, but missing private_key_file");
+		if (!!cfg->certificate_file != !!cfg->private_key_file) {
+			nvnc_log(NVNC_LOG_ERROR, "Need both certificate_file and private_key_file for TLS");
 			rc = -1;
 		}
+
 		if (!cfg->username && !cfg->enable_pam) {
 			nvnc_log(NVNC_LOG_ERROR, "Authentication enabled, but missing username");
 			rc = -1;
@@ -950,6 +1019,11 @@ int check_cfg_sanity(struct cfg* cfg)
 			nvnc_log(NVNC_LOG_ERROR, "Authentication enabled, but missing password");
 			rc = -1;
 		}
+
+		if (cfg->relax_encryption) {
+			nvnc_log(NVNC_LOG_WARNING, "Authentication enabled with relaxed encryption; not all sessions are guaranteed to be encrypted");
+		}
+
 		return rc;
 	}
 
@@ -1000,6 +1074,7 @@ static struct wayvnc_client* client_create(struct wayvnc* wayvnc,
 	self->nvnc_client = nvnc_client;
 
 	self->id = next_client_id++;
+	client_init_seat(self);
 	client_init_keyboard(self);
 	client_init_pointer(self);
 	client_init_data_control(self);
@@ -1010,6 +1085,30 @@ static struct wayvnc_client* client_create(struct wayvnc* wayvnc,
 static void client_destroy(void* obj)
 {
 	struct wayvnc_client* self = obj;
+	struct nvnc* nvnc = nvnc_client_get_server(self->nvnc_client);
+	struct wayvnc* wayvnc = nvnc_get_userdata(nvnc);
+
+	if (self->seat)
+		self->seat->occupancy--;
+
+	wayvnc->nr_clients--;
+	nvnc_log(NVNC_LOG_DEBUG, "Client disconnected, new client count: %d",
+			wayvnc->nr_clients);
+
+	struct ctl_server_client_info info = {
+		.id = self->id,
+		.hostname = nvnc_client_get_hostname(self->nvnc_client),
+		.username = nvnc_client_get_auth_username(self->nvnc_client),
+		.seat = self->seat ? self->seat->name : NULL,
+	};
+
+	ctl_server_event_disconnected(wayvnc->ctl, &info, wayvnc->nr_clients);
+
+	if (wayvnc->nr_clients == 0) {
+		nvnc_log(NVNC_LOG_INFO, "Stopping screen capture");
+		screencopy_stop(&wayvnc->screencopy);
+		stop_performance_ticker(wayvnc);
+	}
 
 	if (self->keyboard.virtual_keyboard) {
 		zwp_virtual_keyboard_v1_destroy(
@@ -1026,30 +1125,6 @@ static void client_destroy(void* obj)
 	free(self);
 }
 
-static void on_nvnc_client_cleanup(struct nvnc_client* client)
-{
-	struct wayvnc_client* wayvnc_client = nvnc_get_userdata(client);
-	struct nvnc* nvnc = nvnc_client_get_server(client);
-	struct wayvnc* self = nvnc_get_userdata(nvnc);
-
-	self->nr_clients--;
-	nvnc_log(NVNC_LOG_DEBUG, "Client disconnected, new client count: %d",
-			self->nr_clients);
-
-	char id[64];
-	snprintf(id, sizeof(id), "%u", wayvnc_client->id);
-	ctl_server_event_disconnected(self->ctl, id,
-			nvnc_client_get_hostname(client),
-			nvnc_client_get_auth_username(client),
-			self->nr_clients);
-
-	if (self->nr_clients == 0) {
-		nvnc_log(NVNC_LOG_INFO, "Stopping screen capture");
-		screencopy_stop(&self->screencopy);
-		stop_performance_ticker(self);
-	}
-}
-
 static void on_nvnc_client_new(struct nvnc_client* client)
 {
 	struct nvnc* nvnc = nvnc_client_get_server(client);
@@ -1065,16 +1140,17 @@ static void on_nvnc_client_new(struct nvnc_client* client)
 		wayvnc_start_capture_immediate(self);
 	}
 	self->nr_clients++;
-	nvnc_set_client_cleanup_fn(client, on_nvnc_client_cleanup);
 	nvnc_log(NVNC_LOG_DEBUG, "Client connected, new client count: %d",
 			self->nr_clients);
 
-	char id[64];
-	snprintf(id, sizeof(id), "%u", wayvnc_client->id);
-	ctl_server_event_connected(self->ctl, id,
-			nvnc_client_get_hostname(client),
-			nvnc_client_get_auth_username(client),
-			self->nr_clients);
+	struct ctl_server_client_info info = {
+		.id = wayvnc_client->id,
+		.hostname = nvnc_client_get_hostname(client),
+		.username = nvnc_client_get_auth_username(client),
+		.seat = wayvnc_client->seat ? wayvnc_client->seat->name : NULL,
+	};
+
+	ctl_server_event_connected(self->ctl, &info, self->nr_clients);
 }
 
 void parse_keyboard_option(struct wayvnc* self, const char* arg)
@@ -1107,16 +1183,36 @@ static void client_init_pointer(struct wayvnc_client* self)
 
 	self->pointer.pointer = pointer_manager_version >= 2
 		? zwlr_virtual_pointer_manager_v1_create_virtual_pointer_with_output(
-			wayvnc->pointer_manager, wayvnc->selected_seat->wl_seat,
+			wayvnc->pointer_manager, self->seat->wl_seat,
 			wayvnc->selected_output->wl_output)
 		: zwlr_virtual_pointer_manager_v1_create_virtual_pointer(
-			wayvnc->pointer_manager, wayvnc->selected_seat->wl_seat);
+			wayvnc->pointer_manager, self->seat->wl_seat);
 
 	if (pointer_init(&self->pointer) < 0) {
 		nvnc_log(NVNC_LOG_ERROR, "Failed to initialise pointer");
 	}
 }
 
+static void client_init_seat(struct wayvnc_client* self)
+{
+	struct wayvnc* wayvnc = self->server;
+
+	if (wayvnc->disable_input)
+		return;
+
+	if (wayvnc->selected_seat) {
+		self->seat = wayvnc->selected_seat;
+	} else {
+		self->seat = seat_find_unoccupied(&wayvnc->seats);
+		if (!self->seat) {
+			self->seat = seat_first(&wayvnc->seats);
+		}
+	}
+
+	if (self->seat)
+		self->seat->occupancy++;
+}
+
 static void client_init_keyboard(struct wayvnc_client* self)
 {
 	struct wayvnc* wayvnc = self->server;
@@ -1126,8 +1222,7 @@ static void client_init_keyboard(struct wayvnc_client* self)
 
 	self->keyboard.virtual_keyboard =
 		zwp_virtual_keyboard_manager_v1_create_virtual_keyboard(
-			wayvnc->keyboard_manager,
-			wayvnc->selected_seat->wl_seat);
+			wayvnc->keyboard_manager, self->seat->wl_seat);
 
 	struct xkb_rule_names rule_names = {
 		.rules = wayvnc->cfg.xkb_rules,
@@ -1163,7 +1258,7 @@ static void client_init_data_control(struct wayvnc_client* self)
 
 	self->data_control.manager = wayvnc->data_control_manager;
 	data_control_init(&self->data_control, wayvnc->display, wayvnc->nvnc,
-			wayvnc->selected_seat->wl_seat);
+			self->seat->wl_seat);
 }
 
 void log_selected_output(struct wayvnc* self)
@@ -1256,6 +1351,7 @@ int main(int argc, char* argv[])
 	const char* address = NULL;
 	int port = 0;
 	bool use_unix_socket = false;
+	bool use_websocket = false;
 
 	const char* output_name = NULL;
 	const char* seat_name = NULL;
@@ -1305,6 +1401,8 @@ int main(int argc, char* argv[])
 		  "Show version info." },
 		{ 'v', "verbose", NULL,
 		  "Be more verbose. Same as setting --log-level=info" },
+		{ 'w', "websocket", NULL,
+		  "Create a websocket." },
 		{ 'L', "log-level", "<level>",
 		  "Set log level. The levels are: error, warning, info, debug trace and quiet.",
 		  .default_ = "warning" },
@@ -1335,7 +1433,8 @@ int main(int argc, char* argv[])
 	socket_path = option_parser_get_value(&option_parser, "socket");
 	overlay_cursor = !!option_parser_get_value(&option_parser, "render-cursor");
 	show_performance = !!option_parser_get_value(&option_parser, "performance");
-	use_unix_socket = !!option_parser_get_value(&option_parser, "unix");
+	use_unix_socket = !!option_parser_get_value(&option_parser, "unix-socket");
+	use_websocket = !!option_parser_get_value(&option_parser, "websocket");
 	disable_input = !!option_parser_get_value(&option_parser, "disable-input");
 	log_level = option_parser_get_value(&option_parser, "verbose")
 		? NVNC_LOG_INFO : NVNC_LOG_WARNING;
@@ -1361,6 +1460,11 @@ int main(int argc, char* argv[])
 		return 1;
 	}
 
+	if (use_unix_socket && use_websocket) {
+		nvnc_log(NVNC_LOG_ERROR, "websocket and unix-socket are conflicting options");
+		return 1;
+	}
+
 	errno = 0;
 	int cfg_rc = cfg_load(&self.cfg, cfg_file);
 	if (cfg_rc != 0 && (cfg_file || errno != ENOENT)) {
@@ -1415,12 +1519,6 @@ int main(int argc, char* argv[])
 			nvnc_log(NVNC_LOG_ERROR, "No such seat");
 			goto failure;
 		}
-	} else if (!self.disable_input) {
-		seat = seat_first(&self.seats);
-		if (!seat) {
-			nvnc_log(NVNC_LOG_ERROR, "No seat found");
-			goto failure;
-		}
 	}
 
 	self.selected_seat = seat;
@@ -1445,7 +1543,13 @@ int main(int argc, char* argv[])
 	if (init_main_loop(&self) < 0)
 		goto main_loop_failure;
 
-	if (init_nvnc(&self, address, port, use_unix_socket) < 0)
+	enum socket_type socket_type = SOCKET_TYPE_TCP;
+	if (use_unix_socket)
+		socket_type = SOCKET_TYPE_UNIX;
+	else if (use_websocket)
+		socket_type = SOCKET_TYPE_WEBSOCKET;
+
+	if (init_nvnc(&self, address, port, socket_type) < 0)
 		goto nvnc_failure;
 
 	if (self.screencopy.manager)
@@ -1466,7 +1570,8 @@ int main(int argc, char* argv[])
 		.userdata = &self,
 		.on_output_cycle = on_output_cycle,
 		.on_output_switch = on_output_switch,
-		.get_client_list = get_client_list,
+		.client_next = client_next,
+		.client_info = client_info,
 		.get_output_list = get_output_list,
 		.on_disconnect_client = on_disconnect_client,
 		.on_wayvnc_exit = on_wayvnc_exit,
