@@ -19,6 +19,8 @@
 #include <stdlib.h>
 #include <string.h>
 #include <stdbool.h>
+#include <errno.h>
+#include <assert.h>
 #include <wayland-client-protocol.h>
 #include <wayland-client.h>
 #include <neatvnc.h>
@@ -27,6 +29,10 @@
 #include "strlcpy.h"
 
 #include "xdg-output-unstable-v1.h"
+#include "wlr-output-power-management-unstable-v1.h"
+
+extern struct zxdg_output_manager_v1* xdg_output_manager;
+extern struct zwlr_output_power_manager_v1* wlr_output_power_manager;
 
 #define MIN(a, b) ((a) < (b) ? (a) : (b))
 #define MAX(a, b) ((a) > (b) ? (a) : (b))
@@ -178,7 +184,10 @@ static const struct wl_output_listener output_listener = {
 
 void output_destroy(struct output* output)
 {
-	zxdg_output_v1_destroy(output->xdg_output);
+	if (output->xdg_output)
+		zxdg_output_v1_destroy(output->xdg_output);
+	if (output->wlr_output_power)
+		zwlr_output_power_v1_destroy(output->wlr_output_power);
 	wl_output_destroy(output->wl_output);
 	free(output);
 }
@@ -194,23 +203,6 @@ void output_list_destroy(struct wl_list* list)
 	}
 }
 
-struct output* output_new(struct wl_output* wl_output, uint32_t id)
-{
-	struct output* output = calloc(1, sizeof(*output));
-	if (!output) {
-		nvnc_log(NVNC_LOG_ERROR, "OOM");
-		return NULL;
-	}
-
-	output->wl_output = wl_output;
-	output->id = id;
-
-	wl_output_add_listener(output->wl_output, &output_listener,
-			output);
-
-	return output;
-}
-
 void output_logical_position(void* data, struct zxdg_output_v1* xdg_output,
                              int32_t x, int32_t y)
 {
@@ -227,6 +219,7 @@ void output_name(void* data, struct zxdg_output_v1* xdg_output,
 	struct output* self = data;
 
 	strlcpy(self->name, name, sizeof(self->name));
+	nvnc_trace("Output %u name: %s", self->id, self->name);
 }
 
 void output_description(void* data, struct zxdg_output_v1* xdg_output,
@@ -235,6 +228,7 @@ void output_description(void* data, struct zxdg_output_v1* xdg_output,
 	struct output* self = data;
 
 	strlcpy(self->description, description, sizeof(self->description));
+	nvnc_trace("Output %u description: %s", self->id, self->description);
 }
 
 static const struct zxdg_output_v1_listener xdg_output_listener = {
@@ -245,13 +239,97 @@ static const struct zxdg_output_v1_listener xdg_output_listener = {
 	.description = output_description,
 };
 
-void output_set_xdg_output(struct output* self,
-                           struct zxdg_output_v1* xdg_output)
+static void output_setup_xdg_output_manager(struct output* self)
 {
-	self->xdg_output = xdg_output;
+	if (!xdg_output_manager || self->xdg_output)
+		return;
 
+	struct zxdg_output_v1* xdg_output =
+		zxdg_output_manager_v1_get_xdg_output(
+			xdg_output_manager, self->wl_output);
+	self->xdg_output = xdg_output;
 	zxdg_output_v1_add_listener(self->xdg_output, &xdg_output_listener,
-	                            self);
+			self);
+}
+
+const char* output_power_state_name(enum output_power_state state)
+{
+	switch(state) {
+	case OUTPUT_POWER_ON:
+		return "ON";
+	case OUTPUT_POWER_OFF:
+		return "OFF";
+	case OUTPUT_POWER_UNKNOWN:
+		return "UNKNOWN";
+	}
+	abort();
+	return NULL;
+}
+
+static void output_power_mode(void *data,
+		     struct zwlr_output_power_v1 *zwlr_output_power_v1,
+		     uint32_t mode)
+{
+	struct output* self = data;
+	nvnc_trace("Output %s power state changed to %s", self->name,
+			(mode == ZWLR_OUTPUT_POWER_V1_MODE_ON) ? "ON" : "OFF");
+
+	enum output_power_state old = self->power;
+	switch (mode) {
+	case ZWLR_OUTPUT_POWER_V1_MODE_OFF:
+		self->power = OUTPUT_POWER_OFF;
+		break;
+	case ZWLR_OUTPUT_POWER_V1_MODE_ON:
+		self->power = OUTPUT_POWER_ON;
+		break;
+	}
+	if (old != self->power && self->on_power_change)
+		self->on_power_change(self);
+}
+
+static void output_power_failed(void *data,
+		     struct zwlr_output_power_v1 *zwlr_output_power_v1)
+{
+	struct output* self = data;
+	nvnc_log(NVNC_LOG_WARNING, "Output %s power state failure", self->name);
+	self->power = OUTPUT_POWER_UNKNOWN;
+	zwlr_output_power_v1_destroy(self->wlr_output_power);
+	self->wlr_output_power = NULL;
+}
+
+static const struct zwlr_output_power_v1_listener wlr_output_power_listener = {
+	.mode = output_power_mode,
+	.failed = output_power_failed,
+};
+
+static void output_setup_wlr_output_power_manager(struct output* self)
+{
+	if (!wlr_output_power_manager || self->wlr_output_power)
+		return;
+
+	struct zwlr_output_power_v1* wlr_output_power =
+		zwlr_output_power_manager_v1_get_output_power(
+				wlr_output_power_manager,
+				self->wl_output);
+	self->wlr_output_power = wlr_output_power;
+
+	zwlr_output_power_v1_add_listener(self->wlr_output_power,
+			&wlr_output_power_listener, self);
+}
+
+int output_set_power_state(struct output* output, enum output_power_state state)
+{
+	assert(state != OUTPUT_POWER_UNKNOWN);
+	if (!output->wlr_output_power) {
+		errno = ENOENT;
+		return -1;
+	}
+	nvnc_trace("Output %s requesting power %s", output->name,
+			output_power_state_name(state));
+	int mode = (state == OUTPUT_POWER_ON) ? ZWLR_OUTPUT_POWER_V1_MODE_ON :
+		ZWLR_OUTPUT_POWER_V1_MODE_OFF;
+	zwlr_output_power_v1_set_mode(output->wlr_output_power, mode);
+	return 0;
 }
 
 struct output* output_find_by_id(struct wl_list* list, uint32_t id)
@@ -285,3 +363,50 @@ struct output* output_first(struct wl_list* list)
 
 	return output;
 }
+
+struct output* output_cycle(const struct wl_list* list,
+		const struct output* current,
+		enum output_cycle_direction direction)
+{
+	const struct wl_list* iter = current ? &current->link : list;
+	iter = (direction == OUTPUT_CYCLE_FORWARD) ?
+		iter->next : iter->prev;
+	if (iter == list) {
+		if (wl_list_empty(list))
+			return NULL;
+		iter = (direction == OUTPUT_CYCLE_FORWARD) ?
+			iter->next : iter->prev;
+	}
+	struct output* output;
+	return wl_container_of(iter, output, link);
+}
+
+void output_setup_wl_managers(struct wl_list* list)
+{
+	struct output* output;
+	wl_list_for_each(output, list, link) {
+		output_setup_xdg_output_manager(output);
+		output_setup_wlr_output_power_manager(output);
+	}
+}
+
+struct output* output_new(struct wl_output* wl_output, uint32_t id)
+{
+	struct output* output = calloc(1, sizeof(*output));
+	if (!output) {
+		nvnc_log(NVNC_LOG_ERROR, "OOM");
+		return NULL;
+	}
+
+	output->wl_output = wl_output;
+	output->id = id;
+	output->power = OUTPUT_POWER_UNKNOWN;
+
+	wl_output_add_listener(output->wl_output, &output_listener,
+			output);
+
+	output_setup_xdg_output_manager(output);
+	output_setup_wlr_output_power_manager(output);
+
+	return output;
+}
